import importlib
from typing import TYPE_CHECKING, Annotated, TypeVar

from ovld import Medley, call_next, ovld, recurse

from .ctx import Context
from .typetags import make_tag

#############
# Constants #
#############

if TYPE_CHECKING:  # pragma: no cover
    T = TypeVar("T")
    TaggedSubclass = Annotated[T, None]
else:
    TaggedSubclass = make_tag("TaggedSubclass", 1)


###################
# Implementations #
###################


def _resolve(ref, base):
    if ref is None:
        return base

    if (ncolon := ref.count(":")) == 0:
        mod_name = base.__module__
        symbol = ref
    elif ncolon == 1:
        mod_name, symbol = ref.split(":")
    else:
        raise Exception(f"Bad format for class reference: {ref}")
    try:
        mod = importlib.import_module(mod_name)
        return getattr(mod, symbol)
    except (ModuleNotFoundError, AttributeError) as exc:
        raise Exception(str(exc))


class TaggedSubclassFeature(Medley):
    @ovld(priority=10)
    def serialize(self, t: type[TaggedSubclass], obj: object, ctx: Context, /):
        base = t.pushdown()
        if not isinstance(obj, base):
            raise TypeError(f"'{obj}' is not a subclass of '{base}'")
        objt = type(obj)
        qn = objt.__qualname__
        assert "." not in qn, "Only top-level symbols can be serialized"
        mod = objt.__module__
        rval = call_next(objt, obj, ctx)
        rval["class"] = f"{mod}:{qn}"
        return rval

    def deserialize(self, t: type[TaggedSubclass], obj: dict, ctx: Context, /):
        base = t.pushdown()
        obj = dict(obj)
        cls_name = obj.pop("class", None)
        actual_class = _resolve(cls_name, base)
        return recurse(actual_class, obj, ctx)



import inspect
from datetime import date, datetime, timedelta

import pytest
from ovld import Medley

from serieux import Serieux, serialize
from serieux.ctx import AccessPath, Context
from serieux.exc import ValidationError

from .common import Color, Level, Point, has_312_features, one_test_per_assert


@one_test_per_assert
def test_serialize_scalars():
    assert serialize(0) == 0
    assert serialize(12) == 12
    assert serialize(-3.25) == -3.25
    assert serialize("flagada") == "flagada"
    assert serialize(True) is True
    assert serialize(False) is False
    assert serialize(None) is None


def test_serialize_point():
    pt = Point(1, 2)
    assert serialize(Point, pt) == {"x": 1, "y": 2}


SEP = """
======
"""


def getcodes(fn, *sigs):
    sigs = [(sig if isinstance(sig, tuple) else (sig,)) for sig in sigs]
    codes = [inspect.getsource(fn.resolve(*sig)) for sig in sigs]
    return SEP.join(codes)


def test_point_codegen(file_regression):
    code = getcodes(serialize, (type[Point], Point, Context))
    file_regression.check(code)


def test_serialize_list_of_points():
    pts = [Point(1, 2), Point(3, 4)]
    assert serialize(list[Point], pts) == [
        {"x": 1, "y": 2},
        {"x": 3, "y": 4},
    ]


def test_serialize_dict_of_points():
    pts = {"p1": Point(1, 2), "p2": Point(3, 4)}
    assert serialize(dict[str, Point], pts) == {
        "p1": {"x": 1, "y": 2},
        "p2": {"x": 3, "y": 4},
    }


@has_312_features
def test_serialize_tree():
    from .definitions_py312 import Tree

    tree = Tree(
        left=Tree(
            left=1,
            right=Tree(left=Tree(left=2, right=3), right=Tree(left=4, right=5)),
        ),
        right=Tree(left=Tree(left=6, right=7), right=8),
    )
    assert serialize(Tree[int], tree) == {
        "left": {
            "left": 1,
            "right": {
                "left": {"left": 2, "right": 3},
                "right": {"left": 4, "right": 5},
            },
        },
        "right": {"left": {"left": 6, "right": 7}, "right": 8},
    }


class Special(Medley):
    def serialize(self, typ: type[int], value: int, ctx: Context):
        return value * 10

    def serialize(self, typ: type[int], value: str, ctx: Context):
        return value * 2


def test_override():
    ss = (Serieux + Special)()
    assert ss.serialize(int, 3) == 30
    assert ss.serialize(int, "quack") == "quackquack"
    assert ss.serialize(list[int], [1, 2, 3]) == [10, 20, 30]
    assert ss.serialize(list[int], [1, "2", 3]) == [10, "22", 30]
    assert ss.serialize(Point, Point(8, 9)) == {"x": 80, "y": 90}
    assert ss.serialize(3) == 30


def test_special_serializer_codegen(file_regression):
    custom = (Serieux + Special)()
    code = getcodes(custom.serialize, (type[Point], Point, Context))
    file_regression.check(code)


class quirkint(int):
    pass


class Quirky(Medley):
    def serialize(self, typ: type[int], value: quirkint, ctx: Context):
        return value * 10


def test_override_quirkint():
    ss = (Serieux + Quirky)()
    assert ss.serialize(int, 3) == 3
    assert ss.serialize(int, quirkint(3)) == 30
    assert ss.serialize(Point, Point(8, 9)) == {"x": 8, "y": 9}
    assert ss.serialize(Point, Point(quirkint(8), 9)) == {"x": 80, "y": 9}


class ExtraWeight(Context):
    weight: int


class WeightedImpl(Medley):
    def serialize(self, typ: type[int], value: int, ctx: ExtraWeight):
        return value + ctx.weight


def test_override_state():
    ss = (Serieux + WeightedImpl)()
    assert ss.serialize(int, 3) == 3
    assert ss.serialize(int, 3, ExtraWeight(10)) == 13
    assert ss.serialize(Point, Point(7, 8)) == {"x": 7, "y": 8}
    assert ss.serialize(Point, Point(7, 8), ExtraWeight(10)) == {"x": 17, "y": 18}


def test_serialize_enum():
    assert serialize(Color, Color.RED) == "red"
    assert serialize(list[Color], [Color.GREEN, Color.BLUE, Color.GREEN]) == [
        "green",
        "blue",
        "green",
    ]


def test_serialize_enum_int():
    assert serialize(Level, Level.MED) == 1
    assert serialize(list[Level], [Level.HI, Level.LO, Level.HI]) == [2, 0, 2]


def test_serialize_date():
    assert serialize(date, date(2023, 5, 15)) == "2023-05-15"
    assert serialize(list[date], [date(2023, 5, 15), date(2024, 1, 1)]) == [
        "2023-05-15",
        "2024-01-01",
    ]


def test_serialize_datetime():
    assert serialize(datetime, datetime(2023, 5, 15, 12, 30, 45)) == "2023-05-15T12:30:45"
    assert serialize(
        list[datetime], [datetime(2023, 5, 15, 12, 30, 45), datetime(2024, 1, 1, 0, 0, 0)]
    ) == ["2023-05-15T12:30:45", "2024-01-01T00:00:00"]


def test_serialize_timedelta():
    assert serialize(timedelta, timedelta(seconds=42)) == "42s"
    assert serialize(timedelta, timedelta(seconds=42, microseconds=500000)) == "42500000us"
    assert serialize(timedelta, timedelta(seconds=-10)) == "-10s"
    assert serialize(list[timedelta], [timedelta(seconds=30), timedelta(days=5)]) == [
        "30s",
        "432000s",
    ]


###############
# Error tests #
###############


def test_error_basic():
    with pytest.raises(
        ValidationError, match=r"Cannot serialize object of type 'str' into expected type 'int'"
    ):
        serialize(int, "oh no")


def test_error_dataclass():
    with pytest.raises(
        ValidationError, match=r"Cannot serialize object of type 'str' into expected type 'int'"
    ):
        serialize(Point, Point(x=1, y="oops"), AccessPath())


@has_312_features
def test_error_serialize_tree():
    from .definitions_py312 import Tree

    tree = Tree(Tree("a", 2), "b")

    with pytest.raises(ValidationError, match=r"At path \.left\.right"):
        serialize(Tree[str], tree, AccessPath())


def test_error_serialize_list():
    li = [0, 1, 2, 3, "oops", 5, 6]

    with pytest.raises(ValidationError, match=r"At path .4"):
        serialize(list[int], li, AccessPath())


def test_error_serialize_list_of_lists():
    li = [[0, 1], [2, 3, "oops", 5, 6]]

    with pytest.raises(ValidationError, match=r"At path .1.2"):
        serialize(list[list[int]], li, AccessPath())



from functools import cached_property

from ovld import Medley, call_next, ovld

from .ctx import Context
from .typetags import NewTag

Lazy = NewTag["Lazy", 1, False]
DeepLazy = NewTag["DeepLazy", 1]


class LazyProxy:
    def __init__(self, evaluate):
        self._evaluate = evaluate

    @cached_property
    def _obj(self):
        return self._evaluate()

    def __getattribute__(self, name):
        if name in ("_obj", "_evaluate", "__dict__"):
            return object.__getattribute__(self, name)
        return getattr(self._obj, name)

    def __str__(self):
        return str(self._obj)

    def __repr__(self):
        return repr(self._obj)

    def __eq__(self, other):
        return self._obj == other

    def __hash__(self):
        return hash(self._obj)

    def __len__(self):
        return len(self._obj)

    def __getitem__(self, key):
        return self._obj[key]

    def __iter__(self):
        return iter(self._obj)

    def __bool__(self):
        return bool(self._obj)

    def __contains__(self, item):
        return item in self._obj

    def __add__(self, other):
        return self._obj + other

    def __sub__(self, other):
        return self._obj - other

    def __mul__(self, other):
        return self._obj * other

    def __truediv__(self, other):
        return self._obj / other

    def __floordiv__(self, other):
        return self._obj // other

    def __mod__(self, other):
        return self._obj % other

    def __pow__(self, other):
        return self._obj**other

    def __radd__(self, other):
        return other + self._obj

    def __rsub__(self, other):
        return other - self._obj

    def __rmul__(self, other):
        return other * self._obj

    def __rtruediv__(self, other):
        return other / self._obj

    def __rfloordiv__(self, other):
        return other // self._obj

    def __rmod__(self, other):
        return other % self._obj

    def __rpow__(self, other):
        return other**self._obj

    def __neg__(self):
        return -self._obj

    def __pos__(self):
        return +self._obj

    def __abs__(self):
        return abs(self._obj)


class LazyDeserialization(Medley):
    @ovld(priority=1)
    def deserialize(self, typ: type[Lazy], value: object, ctx: Context):
        def evaluate():
            return call_next(typ.pushdown(), value, ctx)

        return LazyProxy(evaluate)

    @ovld(priority=1)
    def deserialize(self, typ: type[DeepLazy], value: object, ctx: Context):
        def evaluate():
            return call_next(typ, value, ctx)

        return LazyProxy(evaluate)



import sys
from datetime import date, datetime, timedelta
from pathlib import Path

import pytest

from serieux import deserialize
from serieux.ctx import AccessPath
from serieux.exc import ValidationError

from .common import Color, Defaults, Level, Point, Point3D, has_312_features, one_test_per_assert

here = Path(__file__).parent


@one_test_per_assert
def test_deserialize_scalars():
    assert deserialize(int, 0) == 0
    assert deserialize(int, 12) == 12
    assert deserialize(float, -3.25) == -3.25
    assert deserialize(str, "flagada") == "flagada"
    assert deserialize(bool, True) is True
    assert deserialize(bool, False) is False
    assert deserialize(type(None), None) is None


def test_deserialize_dict():
    assert deserialize(dict[str, int], {"a": 1, "b": 2}) == {"a": 1, "b": 2}


def test_deserialize_point():
    assert deserialize(Point, {"x": 1, "y": 2}) == Point(1, 2)


def test_deserialize_list_of_points():
    pts = [
        {"x": 1, "y": 2},
        {"x": 3, "y": 4},
    ]
    assert deserialize(list[Point], pts) == [Point(1, 2), Point(3, 4)]


def test_deserialize_dict_of_points():
    pts = {
        "pt1": {"x": 1, "y": 2},
        "pt2": {"x": 3, "y": 4},
    }
    assert deserialize(dict[str, Point], pts) == {
        "pt1": Point(1, 2),
        "pt2": Point(3, 4),
    }


@one_test_per_assert
def test_deserialize_union():
    assert deserialize(str | int, 3) == 3
    assert deserialize(str | int, "wow") == "wow"
    assert deserialize(Point | int, 3) == 3


@has_312_features
def test_deserialize_tree():
    from .definitions_py312 import Tree

    tree = {
        "left": {
            "left": 1,
            "right": 2,
        },
        "right": {
            "left": {
                "left": {
                    "left": 3,
                    "right": 4,
                },
                "right": 5,
            },
            "right": 6,
        },
    }

    assert deserialize(Tree[int], tree) == Tree(Tree(1, 2), Tree(Tree(Tree(3, 4), 5), 6))


def test_deserialize_overlapping_union():
    P = Point | Point3D
    assert type(deserialize(P, {"x": 1, "y": 2})) is Point
    assert type(deserialize(P, {"x": 1, "y": 2, "z": 3})) is Point3D

    # Make sure it also works the other way around
    P = Point3D | Point
    assert type(deserialize(P, {"x": 1, "y": 2})) is Point
    assert type(deserialize(P, {"x": 1, "y": 2, "z": 3})) is Point3D


def test_deserialize_defaults():
    data1 = {"name": "bob"}
    data2 = {"cool": True, "name": "alice"}

    x1 = deserialize(Defaults, data1)
    assert not x1.cool

    x2 = deserialize(Defaults, data2)
    assert x2.cool

    assert not x1.aliases
    assert not x2.aliases
    assert x1.aliases is not x2.aliases


def test_deserialize_enum():
    assert deserialize(Color, "red") == Color.RED
    assert deserialize(list[Color], ["green", "blue", "green"]) == [
        Color.GREEN,
        Color.BLUE,
        Color.GREEN,
    ]


def test_deserialize_enum_int():
    assert deserialize(Level, 1) == Level.MED
    assert deserialize(list[Level], [2, 0, 2]) == [Level.HI, Level.LO, Level.HI]


def test_deserialize_date():
    assert deserialize(date, "2023-05-15") == date(2023, 5, 15)
    assert deserialize(list[date], ["2023-05-15", "2024-01-01"]) == [
        date(2023, 5, 15),
        date(2024, 1, 1),
    ]


def test_deserialize_datetime():
    assert deserialize(datetime, "2023-05-15T12:30:45") == datetime(2023, 5, 15, 12, 30, 45)
    assert deserialize(list[datetime], ["2023-05-15T12:30:45", "2024-01-01T00:00:00"]) == [
        datetime(2023, 5, 15, 12, 30, 45),
        datetime(2024, 1, 1, 0, 0, 0),
    ]


def test_deserialize_timedelta():
    assert deserialize(timedelta, "42s") == timedelta(seconds=42)
    assert deserialize(timedelta, "42500000us") == timedelta(seconds=42, microseconds=500000)
    assert deserialize(timedelta, "-10s") == timedelta(seconds=-10)
    assert deserialize(timedelta, "6h30m") == timedelta(hours=6, minutes=30)
    assert deserialize(timedelta, "+8d") == timedelta(days=8)
    assert deserialize(timedelta, "4.5d") == timedelta(days=4, seconds=60 * 60 * 12)
    assert deserialize(timedelta, "1d12h") == timedelta(days=1, hours=12)
    assert deserialize(timedelta, "36h") == timedelta(days=1, hours=12)
    assert deserialize(list[timedelta], ["30s", "5d"]) == [
        timedelta(seconds=30),
        timedelta(days=5),
    ]


###############
# Error tests #
###############


def test_deserialize_scalar_error():
    with pytest.raises(ValidationError, match=r"Cannot deserialize object of type 'str'"):
        deserialize(int, "foo")


def test_deserialize_missing_field():
    pts = [
        {"x": 1, "y": 2},
        {"x": 3},
    ]
    with pytest.raises(ValidationError, match=r"At path .1: KeyError: 'y'"):
        deserialize(list[Point], pts, AccessPath())


def test_error_display(capsys, file_regression):
    pts = [
        {"x": 1, "y": 2},
        {"x": 3},
    ]
    with pytest.raises(ValidationError, match=r"At path .1: KeyError: 'y'") as exc:
        deserialize(list[Point], pts, AccessPath())

    exc.value.display(file=sys.stderr)
    cap = capsys.readouterr()
    file_regression.check("\n".join([cap.out, "=" * 80, cap.err]))



from io import StringIO

import pytest

from serieux.exc import ValidationError, ValidationExceptionGroup


@pytest.hookimpl()
def pytest_exception_interact(node, call, report):
    if call.excinfo.type == ValidationExceptionGroup or call.excinfo.type == ValidationError:
        exc = call.excinfo.value
        io = StringIO()
        exc.display(file=io)
        entry = report.longrepr.reprtraceback.reprentries[-1]
        entry.style = "short"
        content = io.getvalue()
        entry.lines = [content] + [""] * content.count("\n")
        report.longrepr.reprtraceback.reprentries = [entry]



from dataclasses import dataclass, replace
from functools import cache
from types import UnionType
from typing import Union, get_args, get_origin

from ovld import recurse, subclasscheck
from ovld.mro import Order

from .model import Model, model


@dataclass(frozen=True)
class Tag:
    name: str
    priority: int
    inherit: bool = True


def make_tag(name, priority=1, inherit=True) -> type:
    tag = Tag(name=name, priority=priority, inherit=inherit)
    return _create(frozenset({tag}), object)


class NewTag:
    def __class_getitem__(cls, params) -> type:
        if not isinstance(params, tuple):
            params = (params,)
        return make_tag(*params)


@cache
def _create(tags, cls):
    if isinstance(cls, type) and issubclass(cls, TaggedType):
        return _create(tags | cls._tags, cls._cls)
    if not tags:
        return cls
    else:
        name = "&".join(t.name for t in tags)
        clsname = getattr(cls, "__name__", str(cls))
        return type(f"{name}[{clsname}]", (TaggedType,), {"_tags": tags, "_cls": cls})


class TaggedType(type):
    _cls = object
    _tags = frozenset()

    @classmethod
    def __is_supertype__(self, other):
        return (
            isinstance(other, type)
            and issubclass(other, TaggedType)
            and other._tags.issuperset(self._tags)
            and subclasscheck(other._cls, self._cls)
        )

    @classmethod
    def __type_order__(self, other):
        if not (isinstance(other, type) and issubclass(other, TaggedType)):
            return NotImplemented
        prio = tuple(sorted(tag.priority for tag in self._tags))
        prio_o = tuple(sorted(tag.priority for tag in other._tags))
        return Order.LESS if prio > prio_o else Order.MORE if prio < prio_o else Order.NONE

    def __class_getitem__(self, t):
        return _create(self._tags, t)

    @classmethod
    def strip(cls, t):
        if isinstance(t, type) and issubclass(t, TaggedType):
            return _create(t._tags - cls._tags, t._cls)
        return t

    @classmethod
    def pushdown(self):
        return pushdown(self)

    @classmethod
    def transfer(self, t):
        return self[t]


def pushdown(cls):
    if not isinstance(cls, type) or not issubclass(cls, TaggedType):
        return cls
    typ = cls.strip(cls)
    cls = _create(cls._tags - {tag for tag in cls._tags if not tag.inherit}, cls._cls)
    if not isinstance(cls, type) or not issubclass(cls, TaggedType):
        return cls
    if isinstance(typ, type) and issubclass(typ, Model):
        return Model.make(
            original_type=typ.original_type,
            fields=[replace(field, type=cls[field.type]) for field in typ.fields],
            constructor=typ.constructor,
        )
    elif orig := get_origin(typ):
        args = get_args(typ)
        if orig is UnionType:
            orig = Union
        return orig[tuple([cls[a] for a in args])]
    else:
        return typ


def strip_all(cls):
    if isinstance(cls, type) and issubclass(cls, TaggedType):
        return cls.strip(cls)
    return cls


@model.register
def model(t: type[TaggedType]):
    return t[recurse(t._cls)]



from dataclasses import dataclass

import pytest

from serieux.impl import BaseImplementation
from serieux.tsubclass import TaggedSubclass, TaggedSubclassFeature

featured = (BaseImplementation + TaggedSubclassFeature)()
serialize = featured.serialize
deserialize = featured.deserialize


@dataclass
class Animal:
    name: str


@dataclass
class Cat(Animal):
    selfishness: int

    def cry(self):
        return "me" * self.selfishness + "ow"


@dataclass
class Wolf(Animal):
    size: int

    def cry(self):
        "a-woo" + "o" * self.size


def test_tagged_subclass():
    orig = Wolf(name="Wolfie", size=10)
    ser = serialize(TaggedSubclass[Animal], orig)
    assert ser == {
        "class": "tests.test_tsubclass:Wolf",
        "name": "Wolfie",
        "size": 10,
    }
    deser = deserialize(TaggedSubclass[Animal], ser)
    assert deser == orig


def test_wrong_class():
    orig = Wolf(name="Wolfie", size=10)
    with pytest.raises(TypeError, match="Wolf.*is not a subclass of.*Cat"):
        serialize(TaggedSubclass[Cat], orig)


def test_resolve_default():
    ser = {"name": "Kevin"}
    assert deserialize(TaggedSubclass[Animal], ser) == Animal(name="Kevin")


def test_resolve_same_file():
    ser = {"class": "Cat", "name": "Katniss", "selfishness": 3}
    assert deserialize(TaggedSubclass[Animal], ser) == Cat(name="Katniss", selfishness=3)


def test_not_found():
    with pytest.raises(Exception, match="no attribute 'Bloop'"):
        ser = {"class": "Bloop", "name": "Quack"}
        deserialize(TaggedSubclass[Animal], ser)


def test_bad_resolve():
    with pytest.raises(Exception, match="Bad format for class reference"):
        ser = {"class": "x:y:z", "name": "Quack"}
        deserialize(TaggedSubclass[Animal], ser)



from dataclasses import replace

from ovld.medley import ChainAll, Medley


class Context(Medley):
    follow = ChainAll()


class EmptyContext(Context):
    def __add__(self, other):
        return other


class AccessPath(Context):
    access_path: tuple = ()

    def follow(self, objt, obj, field):
        return replace(self, access_path=(*self.access_path, (obj, field)))


empty = EmptyContext()



from dataclasses import MISSING, dataclass, fields
from types import UnionType
from typing import Callable, Union, get_args, get_origin

from ovld import Dataclass, call_next, class_check, ovld, recurse

from .docstrings import get_attribute_docstrings
from .utils import UnionAlias, evaluate_hint

UNDEFINED = object()


@class_check
def Modelizable(t):
    m = model(t)
    return isinstance(m, type) and issubclass(m, Model)


@dataclass(kw_only=True)
class Field:
    name: str
    type: type
    description: str = None
    default: object = MISSING
    default_factory: Callable = MISSING

    argument_name: str | int = UNDEFINED
    property_name: str = UNDEFINED
    serialized_name: str = UNDEFINED

    # Not implemented yet
    flatten: bool = False

    # Meta-variable to store in this field
    metavar: str = None

    def __post_init__(self):
        if self.property_name is UNDEFINED:
            self.property_name = self.name
        if self.argument_name is UNDEFINED:  # pragma: no cover
            self.argument_name = self.name
        if self.serialized_name is UNDEFINED:
            self.serialized_name = self.name

    @property
    def required(self):
        return self.default is MISSING and self.default_factory is MISSING


class Model(type):
    original_type = object
    fields = []
    constructor = None

    @staticmethod
    def make(
        original_type,
        fields,
        constructor,
    ):
        return Model(
            f"Model[{getattr(original_type, '__name__', str(original_type))}]",
            (Model,),
            {
                "original_type": original_type,
                "fields": fields,
                "constructor": constructor,
            },
        )

    def __class_getitem__(cls, t):
        return model(t)

    @classmethod
    def accepts(cls, other):
        ot = cls.original_type
        return issubclass(other, get_origin(ot) or ot)


_model_cache = {}
_premade = {}


def _take_premade(t):
    _model_cache[t] = _premade.pop(t)
    return _model_cache[t]


@ovld(priority=100)
def model(t: type[object]):
    t = evaluate_hint(t)
    if t not in _model_cache:
        _premade[t] = Model.make(
            original_type=t,
            fields=[],
            constructor=None,
        )
        _model_cache[t] = call_next(t)
    return _model_cache[t]


@ovld
def model(dc: type[Dataclass]):
    rval = _take_premade(dc)
    tsub = {}
    constructor = dc
    if (origin := get_origin(dc)) is not None:
        tsub = dict(zip(origin.__type_params__, get_args(dc)))
        constructor = origin

    attributes = get_attribute_docstrings(dc)

    rval.fields = [
        Field(
            name=field.name,
            description=attributes.get(field.name, None),
            type=recurse(evaluate_hint(field.type, dc, None, tsub)),
            default=field.default,
            default_factory=field.default_factory,
            flatten=field.metadata.get("flatten", False),
            metavar=field.metadata.get("serieux_metavar", None),
            argument_name=field.name if field.kw_only else i,
        )
        for i, field in enumerate(fields(constructor))
    ]
    rval.constructor = constructor
    return rval


@ovld
def model(u: type[UnionAlias] | type[UnionType]):
    return Union[tuple(recurse(evaluate_hint(t)) for t in get_args(u))]


@ovld
def model(t: type[object]):
    if (origin := get_origin(t)) is not None:
        return origin[tuple(recurse(evaluate_hint(a)) for a in get_args(t))]
    return t



import importlib
import sys
import typing
from types import GenericAlias, NoneType, UnionType
from typing import (
    ForwardRef,
    TypeVar,
    Union,
    _GenericAlias,
    get_args,
    get_origin,
)

from ovld import parametrized_class_check

UnionAlias = type(Union[int, str])


def clsstring(cls):
    cls = getattr(cls, "original_type", cls)
    if args := typing.get_args(cls):
        origin = typing.get_origin(cls) or cls
        args = ", ".join(map(clsstring, args))
        return f"{origin.__name__}[{args}]"
    else:
        r = repr(cls)
        if r.startswith("<class "):
            return cls.__name__
        else:
            return r


#################
# evaluate_hint #
#################


def evaluate_hint(typ, ctx=None, lcl=None, typesub=None):
    if isinstance(typ, str):
        if ctx is not None and not isinstance(ctx, dict):
            if isinstance(ctx, (GenericAlias, _GenericAlias)):
                origin = get_origin(ctx)
                if hasattr(origin, "__type_params__"):
                    subs = {p: arg for p, arg in zip(origin.__type_params__, get_args(ctx))}
                    typesub = {**subs, **(typesub or {})}
                ctx = origin
            if hasattr(ctx, "__type_params__"):
                lcl = {p.__name__: p for p in ctx.__type_params__}
            ctx = importlib.import_module(ctx.__module__).__dict__
        return evaluate_hint(eval(typ, ctx, lcl), ctx, lcl, typesub)

    elif isinstance(typ, (UnionType, GenericAlias, _GenericAlias)):
        origin = get_origin(typ)
        args = get_args(typ)
        if origin is UnionType:
            origin = Union
        new_args = [evaluate_hint(arg, ctx, lcl, typesub) for arg in args]
        return origin[tuple(new_args)]

    elif isinstance(typ, TypeVar):
        return typesub.get(typ, typ) if typesub else typ

    elif isinstance(typ, ForwardRef):
        if sys.version_info >= (3, 13):
            return typ._evaluate(ctx, lcl, type_params=None, recursive_guard=frozenset())
        else:  # pragma: no cover
            return typ._evaluate(ctx, lcl, recursive_guard=frozenset())

    elif isinstance(typ, type):
        return typ

    else:
        raise TypeError("Cannot evaluate hint:", typ, type(typ))


def _json_type_check(t, bound=object):
    origin = get_origin(t)
    if t is typing.Union:
        return all(_json_type_check(t2) for t2 in get_args(t))
    if not isinstance(origin or t, type) or not issubclass(origin or t, bound):
        return False
    if t in (int, float, str, bool, NoneType):
        return True
    if origin is list:
        (et,) = get_args(t)
        return _json_type_check(et)
    if origin is dict:
        kt, vt = get_args(t)
        return (kt is str) and _json_type_check(vt)
    return False


JSONType = parametrized_class_check(_json_type_check)



import ast
import inspect
import sys
import traceback
from ast import NodeTransformer
from contextlib import contextmanager
from dataclasses import dataclass, field
from enum import Enum
from textwrap import dedent

import pytest
from _pytest.assertion.rewrite import AssertionRewriter

from serieux.exc import ValidationExceptionGroup


@dataclass
class Citizen:
    name: str
    birthyear: int
    hometown: str


@dataclass
class Country:
    languages: list[str]
    capital: str
    population: int
    citizens: list[Citizen]


@dataclass
class World:
    countries: dict[str, Country]


@dataclass
class Point:
    x: int
    y: int


@dataclass
class Point3D(Point):
    z: int


class Color(Enum):
    RED = "red"
    GREEN = "green"
    BLUE = "blue"


class Level(Enum):
    HI = 2
    MED = 1
    LO = 0


@dataclass
class Pig:
    # How pink the pig is
    pinkness: float

    weight: float
    """Weight of the pig, in kilograms"""

    # Is the pig...
    # truly...
    beautiful: bool = True  # ...beautiful?


@dataclass
class Defaults:
    name: str
    aliases: list[str] = field(default_factory=list)
    cool: bool = False


@contextmanager
def validation_errors(msgs):
    try:
        yield
    except ValidationExceptionGroup as veg:
        for pth, msg in msgs.items():
            if not any((exc.access_string() == pth and msg in str(exc)) for exc in veg.exceptions):
                traceback.print_exception(veg)
                raise Exception(f"No exception was raised at {pth} for '{msg}'")


class AssertTransformer(NodeTransformer):
    def visit_FunctionDef(self, node):
        newfns = []
        for i, stmt in enumerate(node.body):
            if not isinstance(stmt, ast.Assert):
                raise Exception("@one_test_per_assert requires all statements to be asserts")
            else:
                newfns.append(
                    ast.FunctionDef(
                        name=f"{node.name}_assert{i + 1}",
                        args=node.args,
                        body=[stmt],
                        decorator_list=node.decorator_list,
                        returns=node.returns,
                    )
                )
        return ast.Module(body=newfns, type_ignores=[])


def one_test_per_assert(fn):
    src = dedent(inspect.getsource(fn))
    filename = inspect.getsourcefile(fn)
    tree = ast.parse(src, filename)
    tree = tree.body[0]
    assert isinstance(tree, ast.FunctionDef)
    tree.decorator_list = []
    new_tree = AssertTransformer().visit(tree)
    ast.fix_missing_locations(new_tree)
    _, lineno = inspect.getsourcelines(fn)
    ast.increment_lineno(new_tree, lineno - 1)
    # Use pytest's assertion rewriter for nicer error messages
    AssertionRewriter(filename, None, None).run(new_tree)
    new_fn = compile(new_tree, filename, "exec")
    glb = fn.__globals__
    exec(new_fn, glb, glb)
    return None


has_312_features = pytest.mark.skipif(
    sys.version_info < (3, 12), reason="This test relies on Python 3.12+ syntax"
)



from dataclasses import dataclass, fields
from typing import get_origin

from ovld import Code, Dataclass, ovld, recurse

from .model import Model
from .typetags import TaggedType


class Tell:
    def __lt__(self, other):
        return self.cost() < other.cost()

    def cost(self):
        return 1


@dataclass(frozen=True)
class TypeTell(Tell):
    t: type

    def gen(self, arg):
        return Code("isinstance($arg, $t)", arg=arg, t=self.t)


@dataclass(frozen=True)
class KeyTell(Tell):
    key: str

    def gen(self, arg):
        return Code("(isinstance($arg, dict) and $k in $arg)", arg=arg, k=self.key)

    def cost(self):
        return 2


@ovld
def tells(typ: type[int] | type[str] | type[bool] | type[float] | type[list] | type[dict]):
    return {TypeTell(typ)}


@ovld
def tells(dc: type[Dataclass]):  # pragma: no cover
    # Usually goes through Model
    dc = get_origin(dc) or dc
    return {TypeTell(dict)} | {KeyTell(f.name) for f in fields(dc)}


@ovld
def tells(m: type[Model]):
    return {TypeTell(dict)} | {KeyTell(f.serialized_name) for f in m.fields}


@ovld
def tells(m: type[TaggedType]):
    return recurse(m.pushdown())



import re
from dataclasses import MISSING
from datetime import date, datetime, timedelta
from enum import Enum
from itertools import pairwise
from types import NoneType, UnionType
from typing import Any, get_args, get_origin

from ovld import (
    Code,
    CodegenInProgress,
    CodegenParameter,
    Def,
    Lambda,
    Medley,
    call_next,
    code_generator,
    ovld,
    recurse,
)
from ovld.types import All

from .ctx import Context, empty
from .exc import ValidationError
from .model import Modelizable, model
from .schema import AnnotatedSchema, Schema
from .tell import tells as get_tells
from .typetags import TaggedType, strip_all
from .utils import UnionAlias, clsstring


class BaseImplementation(Medley):
    validate_serialize: CodegenParameter[bool] = True
    validate_deserialize: CodegenParameter[bool] = True

    def __post_init__(self):
        self._schema_cache = {}

    ##################
    # Global helpers #
    ##################

    @classmethod
    def subcode(
        cls, method_name, t, accessor, ctx_t, ctx_expr=Code("$ctx"), after=None, validate=None
    ):
        accessor = accessor if isinstance(accessor, Code) else Code(accessor)
        if validate is None:
            validate = getattr(cls, f"validate_{method_name}")
        method = getattr(cls, method_name)
        if ec := getattr(cls, f"{method_name}_embed_condition")(t):
            ec = All[ec]
        if ec is not None:
            try:
                fn = method.resolve(type[t], ec, ctx_t, after=after)
                cg = getattr(fn, "__codegen__", None)
                if cg:
                    body = cg.create_expression([None, t, accessor, ctx_expr])
                    ot = get_origin(t) or t
                    if not validate:
                        return Code(body)
                    else:
                        return Code(
                            "$body if isinstance($accessor, $t) else $recurse($self, $t, $accessor, $ctx_expr)",
                            body=Code(body),
                            accessor=accessor,
                            t=ot,
                            recurse=method,
                            ctx_expr=ctx_expr,
                        )
            except (CodegenInProgress, ValueError):  # pragma: no cover
                # This is important if we are going to inline recursively
                # a type that refers to itself down the line.
                # We currently never do that.
                pass
        return Code(
            "$recurse($self, $t, $accessor, $ctx_expr)",
            t=t,
            accessor=accessor,
            recurse=method,
            ctx_expr=ctx_expr,
        )

    ########################################
    # serialize:  helpers and entry points #
    ########################################

    @classmethod
    def serialize_embed_condition(cls, t):
        if t in (int, str, bool, float, NoneType):
            return t

    @ovld(priority=-100)
    def serialize(self, t: type[object], obj: object, ctx: Context, /):
        raise ValidationError(
            f"Cannot serialize object of type '{clsstring(type(obj))}'"
            f" into expected type '{clsstring(t)}'.",
            ctx=ctx,
        )

    @ovld(priority=-1)
    def serialize(self, t: type[TaggedType], obj: object, ctx: Context, /):
        return recurse(t.pushdown(), obj, ctx)

    def serialize(self, obj: object, /):
        return recurse(type(obj), obj, empty)

    def serialize(self, t: type[object], obj: object, /):
        return recurse(t, obj, empty)

    #########################################
    # deserialize: helpers and entry points #
    #########################################

    deserialize_embed_condition = serialize_embed_condition

    @ovld(priority=-100)
    def deserialize(self, t: type[object], obj: object, ctx: Context, /):
        raise ValidationError(
            f"Cannot deserialize object of type '{clsstring(type(obj))}'"
            f" into expected type '{clsstring(t)}'.",
            ctx=ctx,
        )

    @ovld(priority=-1)
    def deserialize(self, t: type[TaggedType], obj: object, ctx: Context, /):
        return recurse(t.pushdown(), obj, ctx)

    def deserialize(self, obj: object, /):
        return recurse(object, obj, empty)

    def deserialize(self, t: type[object], obj: object, /):
        return recurse(t, obj, empty)

    ####################################
    # schema: helpers and entry points #
    ####################################

    @ovld(priority=1000)
    def schema(self, t: type[object], ctx: Context, /):
        t = model(t)
        if t not in self._schema_cache:
            self._schema_cache[t] = holder = Schema(t)
            result = call_next(t, ctx)
            holder.update(result)
        return self._schema_cache[t]

    def schema(self, t: type[object], /):
        return recurse(t, empty)

    ################################
    # Implementations: basic types #
    ################################

    for T in (int, str, bool, float, NoneType):

        @code_generator
        def serialize(cls, t: type[T], obj: T, ctx: Context, /):
            return Lambda(Code("$obj"))

        @code_generator
        def deserialize(cls, t: type[T], obj: T, ctx: Context, /):
            return Lambda(Code("$obj"))

    def schema(self, t: type[int], ctx: Context, /):
        return {"type": "integer"}

    def schema(self, t: type[float], ctx: Context, /):
        return {"type": "number"}

    def schema(self, t: type[str], ctx: Context, /):
        return {"type": "string"}

    def schema(self, t: type[bool], ctx: Context, /):
        return {"type": "boolean"}

    def schema(self, t: type[NoneType], ctx: Context, /):
        return {"type": "null"}

    ##########################
    # Implementations: lists #
    ##########################

    @classmethod
    def __generic_codegen_list(cls, method, t, obj, ctx):
        (t,) = get_args(t)
        (lt,) = get_args(t) or (object,)
        if hasattr(ctx, "follow"):
            ctx_expr = Code("$ctx.follow($objt, $obj, IDX)", objt=obj)
            return Lambda(
                "[$lbody for IDX, X in enumerate($obj)]",
                lbody=cls.subcode(method, lt, "X", ctx, ctx_expr=ctx_expr),
            )
        else:
            return Lambda("[$lbody for X in $obj]", lbody=cls.subcode(method, lt, "X", ctx))

    @code_generator
    def serialize(cls, t: type[list], obj: list, ctx: Context, /):
        return cls.__generic_codegen_list("serialize", t, obj, ctx)

    @code_generator
    def deserialize(cls, t: type[list], obj: list, ctx: Context, /):
        return cls.__generic_codegen_list("deserialize", t, obj, ctx)

    def schema(self, t: type[list], ctx: Context, /):
        (lt,) = get_args(t)
        return {"type": "array", "items": recurse(lt, ctx)}

    ##########################
    # Implementations: dicts #
    ##########################

    @classmethod
    def __generic_codegen_dict(cls, method, t: type[dict], obj: dict, ctx: Context, /):
        (t,) = get_args(t)
        kt, vt = get_args(t) or (object, object)
        ctx_expr = (
            Code("$ctx.follow($objt, $obj, K)", objt=obj)
            if hasattr(ctx, "follow")
            else Code("$ctx")
        )
        return Lambda(
            "{$kbody: $vbody for K, V in $obj.items()}",
            kbody=cls.subcode(method, kt, "K", ctx, ctx_expr=ctx_expr),
            vbody=cls.subcode(method, vt, "V", ctx, ctx_expr=ctx_expr),
        )

    @code_generator
    def serialize(cls, t: type[dict], obj: dict, ctx: Context, /):
        return cls.__generic_codegen_dict("serialize", t, obj, ctx)

    @code_generator
    def deserialize(cls, t: type[dict], obj: dict, ctx: Context, /):
        return cls.__generic_codegen_dict("deserialize", t, obj, ctx)

    def schema(self, t: type[dict], ctx: Context, /):
        kt, vt = get_args(t)
        return {"type": "object", "additionalProperties": recurse(vt, ctx)}

    ################################
    # Implementations: Modelizable #
    ################################

    @code_generator
    def serialize(cls, t: type[Modelizable], obj: object, ctx: Context, /):
        (t,) = get_args(t)
        t = model(t)
        if not t.accepts(obj):
            return None
        stmts = []
        follow = hasattr(ctx, "follow")
        for i, f in enumerate(t.fields):
            ctx_expr = (
                Code("$ctx.follow($objt, $obj, $fld)", objt=obj, fld=f.name)
                if follow
                else Code("$ctx")
            )
            stmt = Code(
                f"v_{i} = $setter",
                setter=cls.subcode(
                    "serialize", f.type, f"$obj.{f.property_name}", ctx, ctx_expr=ctx_expr
                ),
            )
            stmts.append(stmt)
        final = Code(
            "return {$[,]parts}",
            parts=[
                Code(
                    f"$fname: v_{i}",
                    fname=f.serialized_name,
                )
                for i, f in enumerate(t.fields)
            ],
        )
        stmts.append(final)
        stmts = [
            "try:",
            stmts,
            "except $ValidationError:",
            ["raise"],
            "except Exception as exc:",
            ["raise $ValidationError(exc=exc, ctx=$ctx) from None"],
        ]
        return Def(stmts, ValidationError=ValidationError)

    @code_generator
    def deserialize(cls, t: type[Modelizable], obj: dict, ctx: Context, /):
        (t,) = get_args(t)
        t = model(t)
        follow = hasattr(ctx, "follow")
        stmts = []
        args = []
        for i, f in enumerate(t.fields):
            ctx_expr = (
                Code("$ctx.follow($objt, $obj, $fld)", objt=obj, fld=f.name)
                if follow
                else Code("$ctx")
            )
            processed = cls.subcode(
                "deserialize",
                f.type,
                Code("$obj[$pname]", pname=f.property_name),
                ctx,
                ctx_expr=ctx_expr,
            )
            if f.metavar:
                expr = Code(f.metavar)
            elif f.required:
                expr = processed
            elif f.default is not MISSING:
                expr = Code(
                    "($processed) if $pname in $obj else $dflt",
                    dflt=f.default,
                    pname=f.property_name,
                    processed=processed,
                )
            elif f.default_factory is not MISSING:
                expr = Code(
                    "($processed) if $pname in $obj else $dflt()",
                    dflt=f.default_factory,
                    pname=f.property_name,
                    processed=processed,
                )
            stmt = Code(f"v_{i} = $expr", expr=expr)
            stmts.append(stmt)
            if isinstance(f.argument_name, str):
                arg = f"{f.argument_name}=v_{i}"
            else:
                arg = f"v_{i}"
            args.append(arg)

        final = Code(
            "return $constructor($[,]parts)",
            constructor=t.constructor,
            parts=[Code(a) for a in args],
        )

        stmts.append(final)

        stmts = [
            "try:",
            stmts,
            "except $ValidationError:",
            ["raise"],
            "except Exception as exc:",
            ["raise $ValidationError(exc=exc, ctx=$ctx) from None"],
        ]
        return Def(stmts, ValidationError=ValidationError)

    def schema(self, t: type[Modelizable], ctx: Context, /):
        t = model(t)
        properties = {}
        required = []

        for f in t.fields:
            fsch = recurse(f.type, ctx)
            extra = {}
            if f.description:
                extra["description"] = f.description
            if f.default is not MISSING:
                extra["default"] = f.default
            fsch = fsch if not f.description else AnnotatedSchema(fsch, **extra)
            properties[f.property_name] = fsch
            if f.required:
                required.append(f.property_name)

        return {"type": "object", "properties": properties, "required": required}

    ###########################
    # Implementations: Unions #
    ###########################

    @code_generator
    def serialize(cls, t: type[UnionAlias], obj: Any, ctx: Context, /):
        (t,) = get_args(t)
        o1, *rest = get_args(t)
        code = cls.subcode("serialize", o1, "$obj", ctx, validate=False)
        for opt in rest:
            code = Code(
                "$ocode if isinstance($obj, $sopt) else $code",
                sopt=strip_all(opt),
                ocode=cls.subcode("serialize", opt, "$obj", ctx, validate=False),
                code=code,
            )
        return Lambda(code)

    @code_generator
    def deserialize(cls, t: type[UnionAlias] | type[UnionType], obj: Any, ctx: Context, /):
        (t,) = get_args(t)
        options = get_args(t)
        tells = [get_tells(model(o)) for o in options]
        for tl1, tl2 in pairwise(tells):
            inter = tl1 & tl2
            tl1 -= inter
            tl2 -= inter

        if sum(not tl for tl in tells) > 1:
            raise Exception("Cannot differentiate the possible union members.")

        options = list(zip(tells, options))
        options.sort(key=lambda x: len(x[0]))

        (_, o1), *rest = options

        code = cls.subcode("deserialize", o1, "$obj", ctx)
        for tls, opt in rest:
            code = Code(
                "($ocode if $cond else $code)",
                cond=min(tls).gen(Code("$obj")),
                code=code,
                ocode=cls.subcode("deserialize", opt, "$obj", ctx),
            )
        return Lambda(code)

    def schema(self, t: type[UnionAlias], ctx: Context, /):
        options = get_args(t)
        return {"oneOf": [recurse(opt, ctx) for opt in options]}

    ##########################
    # Implementations: Enums #
    ##########################

    @code_generator
    def serialize(self, t: type[Enum], obj: Enum, ctx: Context, /):
        return Lambda(Code("$obj.value"))

    @code_generator
    def deserialize(self, t: type[Enum], obj: Any, ctx: Context, /):
        (t,) = get_args(t)
        return Lambda(Code("$t($obj)", t=t))

    def schema(self, t: type[Enum], ctx: Context, /):
        return {"enum": [e.value for e in t]}

    ##########################
    # Implementations: Dates #
    ##########################

    @code_generator
    def serialize(self, t: type[date] | type[datetime], obj: date | datetime, ctx: Context, /):
        return Lambda(Code("$obj.isoformat()"))

    @code_generator
    def deserialize(self, t: type[date] | type[datetime], obj: str, ctx: Context, /):
        (t,) = get_args(t)
        return Lambda(Code("$t.fromisoformat($obj)", t=t))

    def schema(self, t: type[date], ctx: Context, /):
        return {"type": "string", "format": "date"}

    def schema(self, t: type[datetime], ctx: Context, /):
        return {"type": "string", "format": "date-time"}

    ##############################
    # Implementations: timedelta #
    ##############################

    def serialize(self, t: type[timedelta], obj: timedelta, ctx: Context):
        """Serialize timedelta as Xs (seconds) or Xus (microseconds)."""
        seconds = int(obj.total_seconds())
        if obj.microseconds:
            return f"{seconds}{obj.microseconds:06}us"
        else:
            return f"{seconds}s"

    def deserialize(self, t: type[timedelta], obj: str, ctx: Context):
        """Deserialize a combination of days, hours, etc. as a timedelta."""
        units = {
            "d": "days",
            "h": "hours",
            "m": "minutes",
            "s": "seconds",
            "ms": "milliseconds",
            "us": "microseconds",
        }
        sign = 1
        if obj.startswith("-"):
            obj = obj[1:]
            sign = -1
        kw = {}
        parts = re.split(string=obj, pattern="([a-z ]+)")
        if parts[-1] != "":
            raise TypeError("timedelta representation must end with a unit")
        for i in range(len(parts) // 2):
            n = parts[i * 2]
            unit = parts[i * 2 + 1].strip()
            if unit not in units:
                raise TypeError(f"'{unit}' is not a valid timedelta unit")
            try:
                kw[units[unit]] = float(n)
            except ValueError as err:
                raise TypeError(f"Could not convert '{n}' ({units[unit]}) to float") from err
        return sign * timedelta(**kw)

    def schema(self, t: type[timedelta], ctx: Context, /):
        return {
            "type": "string",
            "pattern": r"^[+-]?(\d+[dhms]|\d+ms|\d+us)+$",
        }



from numbers import Number

from serieux.model import model

from .common import Pig, Point, has_312_features
from .definitions import Tree


def test_model_cached():
    ptm1 = model(Point)
    ptm2 = model(Point)
    assert ptm1 is ptm2


def test_model_recursive():
    tm = model(Tree)
    fleft = tm.fields[0]
    assert fleft.name == "left"
    assert fleft.type == tm | Number


@has_312_features
def test_model_recursive_parametric():
    from .definitions_py312 import Tree

    tm = model(Tree[int])
    fleft = tm.fields[0]
    assert fleft.name == "left"
    assert fleft.type == tm | int


def test_model_default():
    assert model(int) is int


def test_model_idempotent():
    assert model(model(int)) is model(int)


def test_field_descriptions():
    m = model(Pig)
    p, w, b = m.fields

    assert p.name == "pinkness"
    assert p.description == "How pink the pig is"

    assert w.name == "weight"
    assert w.description == "Weight of the pig, in kilograms"

    assert b.name == "beautiful"
    assert b.description == "Is the pig...\ntruly...\n...beautiful?"



from enum import Enum
from typing import Counter

from ovld import Medley, call_next, recurse

from .model import Model


class Schema(dict):
    def __init__(self, t):
        self.for_type = t
        super().__init__()

    def compile(self, **kwargs):
        return SchemaCompiler(**kwargs)(self)

    def __eq__(self, other):
        return self is other

    def __hash__(self):
        return id(self)


class AnnotatedSchema(dict):
    def __init__(self, parent, **annotations):
        self.parent = parent
        super().__init__(annotations)


class RefPolicy(str, Enum):
    ALWAYS = "always"
    NOREPEAT = "norepeat"
    MINIMAL = "minimal"
    NEVER = "never"


class SchemaCompiler(Medley):
    ref_policy: RefPolicy = RefPolicy.NOREPEAT
    root: bool = True

    def __post_init__(self):
        self.refs = {}
        self.defs = {}
        self.done = set()
        self.name_indexes = Counter()

    def unique_name(self, t: type[Model]):
        return recurse(t.original_type)

    def unique_name(self, t: type):
        name = t.__name__
        idx = self.name_indexes[name]
        self.name_indexes[name] += 1
        if idx > 0:
            name = f"{name}{idx + 1}"
        return name

    def __call__(self, x: object):
        rval = recurse(x, ("#",))
        if self.root:
            rval["$schema"] = "https://json-schema.org/draft/2020-12/schema"
        if self.defs:
            rval["$defs"] = self.defs
        return rval

    def __call__(self, d: dict, pth: tuple):
        return {k: recurse(v, (*pth, k)) for k, v in d.items()}

    def __call__(self, xs: list, pth: tuple):
        return [recurse(x, (*pth, str(i))) for i, x in enumerate(xs)]

    def __call__(self, x: object, pth: tuple):
        return x

    def __call__(self, x: Schema, pth: tuple):
        is_always = self.ref_policy == RefPolicy.ALWAYS
        if x.get("type", "object") != "object":
            return call_next(x, pth)
        elif x in self.refs:
            if x not in self.done and self.ref_policy == RefPolicy.NEVER:
                raise Exception("Recursive schema cannot be compiled without $ref")
            elif x not in self.done or self.ref_policy not in (RefPolicy.NEVER, RefPolicy.MINIMAL):
                return {"$ref": "/".join(self.refs[x])}
            else:
                return call_next(x, pth)
        else:
            if is_always:
                name = self.unique_name(x.for_type)
                pth = ("#", "$defs", name)
            self.refs[x] = pth
            rval = call_next(x, pth)
            if "$ref" not in rval:
                self.done.add(x)
            if is_always:
                self.defs[name] = rval
                return {"$ref": f"#/$defs/{name}"}
            return rval

    def __call__(self, x: AnnotatedSchema, pth: tuple):
        rval = recurse(x.parent, pth)
        return {**rval, **x}



from .ctx import Context
from .exc import ValidationError, ValidationExceptionGroup
from .impl import BaseImplementation
from .lazy import DeepLazy, Lazy, LazyDeserialization, LazyProxy
from .partial import PartialFeature
from .typetags import NewTag, TaggedType
from .version import version as __version__

Serieux = BaseImplementation + PartialFeature + LazyDeserialization
serieux = Serieux()
serialize = serieux.serialize
deserialize = serieux.deserialize
schema = serieux.schema


__all__ = [
    "Context",
    "NewTag",
    "TaggedType",
    "BaseImplementation",
    "serialize",
    "deserialize",
    "schema",
    "Serieux",
    "serieux",
    "ValidationError",
    "ValidationExceptionGroup",
    "Lazy",
    "DeepLazy",
    "LazyProxy",
    "__version__",
]



import json
import tomllib
from dataclasses import dataclass
from pathlib import Path

import yaml
from ovld import dependent_check, ovld, recurse
from ovld.dependent import HasKey

from .ctx import Context
from .partial import PartialFeature, Sources


@dependent_check
def FileSuffix(value: Path, *suffixes):
    return value.exists() and value.suffix in suffixes


@ovld
def parse(path: FileSuffix[".toml"]):
    return tomllib.loads(path.read_text())


@ovld
def parse(path: FileSuffix[".json"]):
    return json.loads(path.read_text())


@ovld
def parse(path: FileSuffix[".yaml", ".yml"]):
    return yaml.safe_load(path.read_text())


@ovld
def parse_with_source(path: FileSuffix[".yaml", ".yml"]):
    return yaml.compose(path.read_text())


class WorkingDirectory(Context):
    origin: Path
    directory: Path


@dataclass
class Location:
    source: str
    start: int
    end: int
    linecols: tuple


class YamlSourceInfo(Context):
    location: Location

    @classmethod
    def extract(cls, node):
        return cls(
            location=Location(
                source=node.start_mark.buffer,
                start=node.start_mark.index,
                end=node.end_mark.index,
                linecols=(
                    (node.start_mark.line, node.start_mark.column),
                    (node.end_mark.line, node.end_mark.column),
                ),
            )
        )


@dependent_check
def ScalarNode(value: yaml.ScalarNode, tag_suffix):
    return value.tag.endswith(tag_suffix)


class FromFileFeature(PartialFeature):
    @ovld(priority=1)
    def deserialize(self, t: type[object], obj: HasKey["$include"], ctx: Context):
        obj = dict(obj)
        return recurse(t, Sources(Path(obj.pop("$include")), obj), ctx)

    def deserialize(self, t: type[object], obj: Path, ctx: Context):
        if isinstance(ctx, WorkingDirectory):
            obj = ctx.directory / obj
        data = parse(obj)
        ctx = ctx + WorkingDirectory(origin=obj, directory=obj.parent)
        try:
            return recurse(t, data, ctx)
        except Exception:
            pass
        data = parse_with_source(obj)
        return recurse(t, data, ctx)

    def deserialize(self, t: type[object], obj: yaml.MappingNode, ctx: Context):
        return recurse(t, {k.value: v for k, v in obj.value}, ctx + YamlSourceInfo.extract(obj))

    def deserialize(self, t: type[object], obj: yaml.SequenceNode, ctx: Context):
        return recurse(t, obj.value, ctx + YamlSourceInfo.extract(obj))

    def deserialize(self, t: type[object], obj: ScalarNode[":str"], ctx: Context):
        return recurse(t, obj.value, ctx + YamlSourceInfo.extract(obj))

    def deserialize(self, t: type[object], obj: ScalarNode[":int"], ctx: Context):
        return recurse(t, int(obj.value), ctx + YamlSourceInfo.extract(obj))

    def deserialize(self, t: type[object], obj: ScalarNode[":float"], ctx: Context):
        return recurse(t, float(obj.value), ctx + YamlSourceInfo.extract(obj))



from dataclasses import dataclass
from datetime import date, datetime, timedelta

import pytest

from serieux import schema as _schema

from .common import Color, Defaults, Pig, Point, has_312_features


def schema(t, root=False, ref_policy="norepeat"):
    return _schema(t).compile(root=root, ref_policy=ref_policy)


def test_schema_int():
    assert schema(int) == {"type": "integer"}


def test_schema_bool():
    assert schema(bool) == {"type": "boolean"}


def test_schema_str():
    assert schema(str) == {"type": "string"}


def test_schema_enum():
    assert schema(Color) == {"enum": ["red", "green", "blue"]}


def test_schema_date():
    assert schema(date) == {"type": "string", "format": "date"}


def test_schema_datetime():
    assert schema(datetime) == {"type": "string", "format": "date-time"}


def test_schema_timedelta():
    assert schema(timedelta) == {
        "type": "string",
        "pattern": r"^[+-]?(\d+[dhms]|\d+ms|\d+us)+$",
    }


def test_schema_list():
    assert schema(list[int]) == {"type": "array", "items": {"type": "integer"}}


def test_schema_dict():
    assert schema(dict[str, float]) == {
        "type": "object",
        "additionalProperties": {"type": "number"},
    }


def test_schema_nested():
    assert schema(dict[str, list[int]]) == {
        "type": "object",
        "additionalProperties": {"type": "array", "items": {"type": "integer"}},
    }


def test_schema_dataclass():
    assert schema(Point) == {
        "type": "object",
        "properties": {"x": {"type": "integer"}, "y": {"type": "integer"}},
        "required": ["x", "y"],
    }


def test_schema_dataclass_2():
    assert schema(Defaults) == {
        "type": "object",
        "properties": {
            "name": {"type": "string"},
            "aliases": {"type": "array", "items": {"type": "string"}},
            "cool": {"type": "boolean"},
        },
        "required": ["name"],
    }


@has_312_features
def test_schema_recursive():
    from .definitions_py312 import Tree

    assert schema(Tree[int]) == {
        "type": "object",
        "properties": {
            "left": {
                "oneOf": [
                    {"$ref": "#"},
                    {"type": "integer"},
                ]
            },
            "right": {
                "oneOf": [
                    {"$ref": "#"},
                    {"type": "integer"},
                ]
            },
        },
        "required": ["left", "right"],
    }


@has_312_features
def test_schema_recursive_policy_always():
    from .definitions_py312 import Tree

    assert schema(Tree[int], root=True, ref_policy="always") == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "$ref": "#/$defs/Tree",
        "$defs": {
            "Tree": {
                "type": "object",
                "properties": {
                    "left": {
                        "oneOf": [
                            {"$ref": "#/$defs/Tree"},
                            {"type": "integer"},
                        ]
                    },
                    "right": {
                        "oneOf": [
                            {"$ref": "#/$defs/Tree"},
                            {"type": "integer"},
                        ]
                    },
                },
                "required": ["left", "right"],
            }
        },
    }


@has_312_features
def test_schema_recursive_policy_never():
    from .definitions_py312 import Tree

    with pytest.raises(Exception, match="Recursive schema"):
        schema(Tree[int], root=True, ref_policy="never")


@dataclass
class TwoPoints:
    # First point
    a: Point
    # Second point
    b: Point


def test_schema_policy_never_minimal():
    never = schema(TwoPoints, ref_policy="never")
    minimal = schema(TwoPoints, ref_policy="minimal")
    assert (
        never
        == minimal
        == {
            "type": "object",
            "properties": {
                "a": {
                    "type": "object",
                    "description": "First point",
                    "properties": {"x": {"type": "integer"}, "y": {"type": "integer"}},
                    "required": ["x", "y"],
                },
                "b": {
                    "type": "object",
                    "description": "Second point",
                    "properties": {"x": {"type": "integer"}, "y": {"type": "integer"}},
                    "required": ["x", "y"],
                },
            },
            "required": ["a", "b"],
        }
    )


def test_schema_policy_norepeat():
    assert schema(TwoPoints, ref_policy="norepeat") == {
        "type": "object",
        "properties": {
            "a": {
                "type": "object",
                "description": "First point",
                "properties": {"x": {"type": "integer"}, "y": {"type": "integer"}},
                "required": ["x", "y"],
            },
            "b": {
                "$ref": "#/properties/a",
                "description": "Second point",
            },
        },
        "required": ["a", "b"],
    }


def test_schema_policy_always():
    assert schema(TwoPoints, ref_policy="always") == {
        "$ref": "#/$defs/TwoPoints",
        "$defs": {
            "Point": {
                "type": "object",
                "properties": {"x": {"type": "integer"}, "y": {"type": "integer"}},
                "required": ["x", "y"],
            },
            "TwoPoints": {
                "type": "object",
                "properties": {
                    "a": {"$ref": "#/$defs/Point", "description": "First point"},
                    "b": {"$ref": "#/$defs/Point", "description": "Second point"},
                },
                "required": ["a", "b"],
            },
        },
    }


def test_schema_descriptions():
    assert schema(Pig) == {
        "type": "object",
        "properties": {
            "pinkness": {"type": "number", "description": "How pink the pig is"},
            "weight": {"type": "number", "description": "Weight of the pig, in kilograms"},
            "beautiful": {
                "type": "boolean",
                "description": "Is the pig...\ntruly...\n...beautiful?",
                "default": True,
            },
        },
        "required": ["pinkness", "weight"],
    }



from dataclasses import field, fields, make_dataclass
from functools import reduce

from ovld import Medley, call_next, ovld, recurse

from .ctx import Context
from .exc import ValidationError, ValidationExceptionGroup
from .model import Model, Modelizable, model
from .typetags import NewTag

#############
# Constants #
#############


Partial = NewTag["Partial"]


class NOT_GIVEN_T:
    pass


NOT_GIVEN = NOT_GIVEN_T()


class PartialBase:
    pass


class Sources:
    def __init__(self, *sources):
        self.sources = sources


@ovld
def partialize(t: type[Modelizable]):
    if issubclass(t, Model):
        return recurse(t.original_type)
    if issubclass(t, PartialBase):
        return t
    m = model(t)
    fields = [(f.name, partialize(f.type), field(default=NOT_GIVEN)) for f in m.fields]
    fields.append(
        ("_serieux_ctx", Context, field(default=NOT_GIVEN, metadata={"serieux_metavar": "$ctx"}))
    )
    dc = make_dataclass(
        cls_name=f"Partial[{t.__name__}]",
        bases=(PartialBase,),
        fields=fields,
        namespace={"_constructor": m.constructor},
    )
    return dc


@ovld(priority=1)
def partialize(t: type[Partial]):
    return recurse(t.pushdown())


@ovld
def partialize(t: object):
    return Partial[t]


###################
# Implementations #
###################


class PartialFeature(Medley):
    @ovld(priority=1)
    def deserialize(self, t: type[Partial[object]], obj: object, ctx: Context, /):
        try:
            return call_next(t, obj, ctx)
        except ValidationError as exc:
            return exc

    def deserialize(self, t: type[object], obj: Sources, ctx: Context, /):
        parts = [recurse(Partial[t], src, ctx) for src in obj.sources]
        rval = instantiate(reduce(merge, parts))
        if isinstance(rval, (ValidationError, ValidationExceptionGroup)):
            raise rval
        return rval


@model.register
def _(p: type[Partial[object]]):
    return call_next(partialize(p.pushdown()))


######################
# Merge partial data #
######################


@ovld
def merge(x: object, y: NOT_GIVEN_T):
    return x


@ovld
def merge(x: NOT_GIVEN_T, y: object):
    return y


@ovld
def merge(x: NOT_GIVEN_T, y: NOT_GIVEN_T):
    return NOT_GIVEN


@ovld
def merge(x: PartialBase, y: PartialBase):
    assert x._constructor is y._constructor
    args = {}
    for f in fields(type(x)):
        xv = getattr(x, f.name)
        yv = getattr(y, f.name)
        args[f.name] = recurse(xv, yv)
    return type(x)(**args)


@ovld
def merge(x: dict, y: dict):
    result = dict(x)
    for k, v in y.items():
        result[k] = recurse(result.get(k, NOT_GIVEN), v)
    return result


@ovld
def merge(x: list, y: list):
    return x + y


@ovld
def merge(x: object, y: object):
    return y


############################
# Instantiate partial data #
############################


@ovld
def instantiate(xs: list):
    rval = []
    errs = []
    for v in xs:
        value = recurse(v)
        if isinstance(value, ValidationError):
            errs.append(value)
        elif isinstance(value, ValidationExceptionGroup):
            errs.extend(value.exceptions)
        else:
            rval.append(value)
    if errs:
        return ValidationExceptionGroup("Some errors occurred", errs)
    return rval


@ovld
def instantiate(xs: dict):
    rval = {}
    errs = []
    for k, v in xs.items():
        if v is NOT_GIVEN:
            continue
        value = recurse(v)
        if isinstance(value, ValidationError):
            errs.append(value)
        elif isinstance(value, ValidationExceptionGroup):
            errs.extend(value.exceptions)
        else:
            rval[k] = value
    if errs:
        return ValidationExceptionGroup("Some errors occurred", errs)
    return rval


@ovld
def instantiate(p: PartialBase):
    dc = p._constructor
    args = recurse({f.name: getattr(p, f.name) for f in fields(dc)})
    if isinstance(args, (ValidationError, ValidationExceptionGroup)):
        return args
    try:
        return dc(**args)
    except Exception as exc:
        return ValidationError(exc=exc, ctx=p._serieux_ctx)


@ovld
def instantiate(x: object):
    return x



import sys
from pathlib import Path

import pytest

from serieux.ctx import AccessPath
from serieux.exc import ValidationError
from serieux.fromfile import FromFileFeature
from serieux.impl import BaseImplementation
from serieux.partial import Sources

from .common import Citizen, Country, World

deserialize = (BaseImplementation + FromFileFeature)().deserialize

here = Path(__file__).parent


def test_deserialize_from_file():
    assert deserialize(Country, here / "data" / "canada.yaml") == Country(
        languages=["English", "French"],
        capital="Ottawa",
        population=39_000_000,
        citizens=[
            Citizen(
                name="Olivier",
                birthyear=1985,
                hometown="Montreal",
            ),
            Citizen(
                name="Abraham",
                birthyear=2018,
                hometown="Shawinigan",
            ),
        ],
    )


def test_deserialize_override():
    srcs = Sources(
        here / "data" / "canada.yaml",
        {"capital": "Montreal"},
    )
    assert deserialize(Country, srcs) == Country(
        languages=["English", "French"],
        capital="Montreal",
        population=39_000_000,
        citizens=[
            Citizen(
                name="Olivier",
                birthyear=1985,
                hometown="Montreal",
            ),
            Citizen(
                name="Abraham",
                birthyear=2018,
                hometown="Shawinigan",
            ),
        ],
    )


def test_deserialize_world():
    world = deserialize(World, here / "data" / "world.yaml")
    assert world == World(
        countries={
            "canada": Country(
                languages=["English", "French"],
                capital="Ottawa",
                population=39_000_000,
                citizens=[
                    Citizen(
                        name="Olivier",
                        birthyear=1985,
                        hometown="Montreal",
                    ),
                    Citizen(
                        name="Abraham",
                        birthyear=2018,
                        hometown="Shawinigan",
                    ),
                ],
            ),
            "france": Country(
                languages=["French"],
                capital="Paris",
                population=68_000_000,
                citizens=[
                    Citizen(
                        name="Jeannot",
                        birthyear=1893,
                        hometown="Lyon",
                    ),
                ],
            ),
        }
    )


def test_deserialize_incomplete(capsys, file_regression):
    with pytest.raises(ValidationError, match="KeyError: 'capital'") as exc:
        deserialize(Country, here / "data" / "france.yaml", AccessPath())

    exc.value.display(file=sys.stderr)
    cap = capsys.readouterr()
    file_regression.check("\n".join([cap.out, "=" * 80, cap.err]))


def test_deserialize_invalid(capsys, file_regression):
    with pytest.raises(ValidationError, match="Cannot deserialize object") as exc:
        deserialize(Country, here / "data" / "invalid.yaml", AccessPath())

    exc.value.display(file=sys.stderr)
    cap = capsys.readouterr()
    file_regression.check("\n".join([cap.out, "=" * 80, cap.err]))


def test_deserialize_oops_world(capsys, file_regression):
    with pytest.raises(ValidationError, match="Cannot deserialize object") as exc:
        deserialize(World, here / "data" / "oops-world.yaml", AccessPath())

    exc.value.display(file=sys.stderr)
    cap = capsys.readouterr()
    file_regression.check("\n".join([cap.out, "=" * 80, cap.err]))



from dataclasses import dataclass

import pytest

from serieux.ctx import AccessPath
from serieux.exc import ValidationError
from serieux.impl import BaseImplementation
from serieux.partial import Partial, PartialFeature, Sources
from tests.common import Point, validation_errors

deserialize = (BaseImplementation + PartialFeature)().deserialize


def test_partial():
    one = deserialize(Partial[Point], {"x": 10})
    assert one.x == 10
    two = deserialize(Partial[Point], {"y": 30})
    assert two.y == 30


def test_partial_error():
    value = deserialize(Partial[Point], {"x": "oh", "y": "no"})
    assert isinstance(value.x, ValidationError)
    assert isinstance(value.y, ValidationError)


def test_two_sources():
    assert deserialize(Point, Sources({"x": 1}, {"y": 2})) == Point(1, 2)


def test_three_sources():
    assert deserialize(Point, Sources({"x": 1}, {"y": 2}, {"x": 3})) == Point(3, 2)


def test_complicated_partial():
    d = deserialize(
        dict[str, Point | str],
        Sources(
            {"a": {"x": 1, "y": 2}},
            {"a": {"x": 3}},
            {"b": "wow"},
        ),
    )
    assert d == {"a": Point(3, 2), "b": "wow"}


@dataclass
class Climate:
    hot: bool
    sunny: bool


@dataclass
class City:
    name: str
    population: int
    climate: Climate


@dataclass
class Country:
    name: str
    capital: City


def test_nested():
    d = deserialize(
        Country,
        Sources(
            {"name": "Canada"},
            {"capital": {"name": "Ottawa"}},
            {"capital": {"population": 800000}},
            {"capital": {"climate": {"hot": False}}},
            {"capital": {"climate": {"sunny": False}}},
        ),
    )
    assert d == Country(
        name="Canada",
        capital=City(
            name="Ottawa",
            population=800_000,
            climate=Climate(
                hot=False,
                sunny=False,
            ),
        ),
    )


@dataclass
class Positive:
    m: int
    n: int

    def __post_init__(self):
        if self.m <= 0:
            raise ValueError("NO")
        if self.n <= 0:
            raise ValueError("Get that worthless number outta here")


def test_error_at_construction():
    with pytest.raises(ValidationError, match="At path .1"):
        deserialize(list[Positive], [{"m": 3, "n": 7}, Sources({"m": 1}, {"n": -3})], AccessPath())


def test_multiple_errors():
    msg = "Cannot deserialize object of type 'str' into expected type 'int'."
    with validation_errors({".0.y": msg, ".1.x": msg, ".1.y": msg}):
        deserialize(
            list[Point], Sources([{"x": 23, "y": "crap"}, {"x": "oh", "y": "no"}]), AccessPath()
        )
